// SHP.cpp: implementation of the CSHP class.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "SHP.h"

#include "MainFrm.h"
#include "FieldDLG.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CSHP::CSHP()
{
	m_bOpen = false;
}

CSHP::~CSHP()
{
	Close();
}

bool CSHP::Open(CString _filename)
{
	FILE* fpSHP = fopen(_filename, "rb");

	// get shp type
	fseek(fpSHP, 32L, SEEK_SET);
	fread(&m_nShapeType, 4, 1, fpSHP);
	
	// data opening..
	switch(m_nShapeType)
	{
	case SHPT_POINT:
		m_pSHPObject = new shpPoint;
		if(!((shpPoint*)m_pSHPObject)->Open(_filename))
			return false;
		break;

	case SHPT_ARC:
		m_pSHPObject = new shpPolyLine;
		if(!((shpPolyLine*)m_pSHPObject)->Open(_filename))
			return false;
		break;

	case SHPT_POLYGON:
		m_pSHPObject = new shpPolygon;
		if(!((shpPolygon*)m_pSHPObject)->Open(_filename))
			return false;
		break;
	}
	fclose(fpSHP);
	m_bOpen = true;

	return true;
}

bool CSHP::Close()
{
	if(m_bOpen)
	{
		switch(m_nShapeType)
		{
		case SHPT_POINT:
			delete (shpPoint*)m_pSHPObject;
			break;

		case SHPT_ARC:
			delete (shpPolyLine*)m_pSHPObject;
			break;
		
		case SHPT_POLYGON:
			delete (shpPolygon*)m_pSHPObject;
			break;
		}
	}
	m_bOpen = false;

	return true;
}

void CSHP::Draw(CDC* _pDC, double _ZoomFactor, GPoint _ptMapCenter, const CRect& _rect, double _lfRatio)
{
	if(m_bOpen)
		m_pSHPObject->Draw(_pDC, _ZoomFactor, _ptMapCenter, _rect, _lfRatio);
}

int	CSHP::Track(CWnd* pWnd, GPoint point)
{
	return m_pSHPObject->Track(pWnd, point);
}

bool CSHP::InfoShow(CWnd* pWnd, GPoint point)
{
	int ID = Track(pWnd, point);
	
	if(ID == -1)
		return false;

	vector<CString> vstrFieldName;	
	vector<CString>	vstrRecord;
 
	// MainFrm ִ ༮ ...
	CInfoDLG*& pInfoDLG = ((CMainFrame*)AfxGetMainWnd())->m_pInfoDLG;
				
	for(int j = 0; j < m_pSHPObject->GetFieldNumber(); ++j)
	{
		char			field[80];
		int				width, decimal;
		CString			tmpString;

		tmpString = m_pSHPObject->ReadStringAttribute(ID, j);
		m_pSHPObject->GetFieldInfo(j, field, &width, &decimal);
		CString tmp(field);

		vstrRecord.push_back(tmpString);
		vstrFieldName.push_back(tmp);
	}

	// 帮 ̾αװ Ǿ     ø
	if (pInfoDLG->GetSafeHwnd() != NULL)
	{
		pInfoDLG->InsertItem(vstrFieldName, vstrRecord);
		pInfoDLG->ShowWindow(SW_SHOW);
		return true;
	}

	pInfoDLG = new CInfoDLG;
	pInfoDLG->Create(134 /* IDD_INFO_DLG */);
	pInfoDLG->InsertItem(vstrFieldName, vstrRecord);
	pInfoDLG->ShowWindow(SW_SHOW);

	return true;
}

void CSHP::SetColor(COLORREF _color)
{
	m_pSHPObject->SetColor(_color);
}

void CSHP::SetBrushColor(COLORREF _color)
{
	m_pSHPObject->SetBrushColor(_color);
}

void CSHP::SetWidth(int _nWidth)
{
	m_pSHPObject->SetWidth(_nWidth);
}

void CSHP::SetStyle(int _nStyle)
{
	m_pSHPObject->SetStyle(_nStyle);
}

void CSHP::SetSolid(bool _bSolid)
{
	m_pSHPObject->SetSolid(_bSolid);
}

void CSHP::ShowLabel()
{
	CFieldDLG	dlg;

	for(int j = 0; j < m_pSHPObject->GetFieldNumber(); ++j)
	{
		char			field[80];
		int				width, decimal;

		m_pSHPObject->GetFieldInfo(j, field, &width, &decimal);
		CString tmp(field);

		dlg.m_vField.push_back(tmp);
	}
	if(dlg.DoModal() != IDOK)
		return;

	if(dlg.m_nSelected != -1)
		m_pSHPObject->SetLabel(dlg.m_nSelected);
}

void CSHP::RemoveLabel()
{
	m_pSHPObject->RemoveLabel();
}

MBR CSHP::GetMBR()
{
	if(m_bOpen)
		return m_pSHPObject->GetMBR();
	else
	{	
		MBR tmpMBR;

		tmpMBR.xmax = tmpMBR.xmin = tmpMBR.ymax = tmpMBR.ymin = 0.;
		return tmpMBR;
	}
}
